from __future__ import annotations
from typing import (
    Type,
    TypeVar,
    Literal,
    Generator,
    Iterator,
    overload,
    Any,
    ClassVar,
)
from types import GenericAlias

import sys
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass, is_dataclass, fields, asdict, InitVar
import traceback


T = TypeVar("T")
ContextDataType = TypeVar("ContextDataType", bound="ContextData")
ContextType = TypeVar("ContextType", bound="Context")


class Context:
    _ctx_var: ContextVar["Context"] = ContextVar("Context")

    @classmethod
    def _decorate_context_type(cls, class_to_decorate):
        # Defaults:
        dataclass_args = {
            "frozen": True,
            "repr": True,
        }
        # Optimizations for python 3.10+
        if sys.version_info >= (3, 10):
            dataclass_args["slots"] = True
            dataclass_args["kw_only"] = True
        decorated = dataclass(class_to_decorate, **dataclass_args)
        return decorated

    @classmethod
    def context_type(cls, class_to_decorate: Type[T]) -> Type[T]:
        if not issubclass(class_to_decorate, ContextData):
            raise TypeError("You must inherit ContextData!")
        for name in class_to_decorate.__annotations__.keys():
            if name in ("_depth", "_on_entering", "_on_exiting", "_on_exited"):
                raise TypeError(
                    f"You cannot declare the ContextData subclass {class_to_decorate} with a field {name!r}: it is a reserved keyword."
                )

        decorated_class = cls._decorate_context_type(class_to_decorate)
        return decorated_class

    @overload
    @classmethod
    def current(
        cls: Type[ContextType], raises: Literal[False]
    ) -> ContextType | None: ...

    @overload
    @classmethod
    def current(cls: Type[ContextType], raises: Literal[True]) -> ContextType: ...

    @overload
    @classmethod
    def current(cls: Type[ContextType]) -> ContextType: ...

    @classmethod
    def current(cls: Type[ContextType], raises: bool = True) -> ContextType | None:
        ctx = cls._ctx_var.get(None)
        if ctx is None:
            if not raises:
                return None
            raise ValueError(f"No context is currently active.")
        if not isinstance(ctx, cls):
            if not raises:
                return None
            raise TypeError(f"The current context is a {ctx.__class__}, not a {cls}.")
        return ctx

    @classmethod
    def _augment(
        cls,
        parent: ContextData | None,
        SubType: Type[ContextDataType],
        kwargs: dict[str, Any],
        skip_extra_fields: bool,
    ) -> ContextDataType:
        sub_kwargs = {}
        if parent is not None:
            parent_fields = fields(parent)  # type: ignore - we know it is a dataclass
            sub_kwargs = {f.name: getattr(parent, f.name) for f in parent_fields}
        sub_kwargs.update(**kwargs)
        if skip_extra_fields:
            filtered = {}
            for field in fields(SubType):  # type: ignore - we know it is a dataclass
                filtered[field.name] = sub_kwargs[field.name]
            sub_kwargs = filtered
        depth = sub_kwargs.get("_depth", 0)
        sub_kwargs["_depth"] = depth + 1
        # TODO: should we validate all field type here?
        # NB: I tried to use pydantic.dataclasses.dataclass for that but it's not that straightforward
        # If we decide that ContextData must validate their fields, we should probably drop dataclasses and use pydantic models...
        try:
            sub = SubType(**sub_kwargs)
        except TypeError as err:
            raise TypeError(
                f"While augmenting a {parent.__class__.__name__} to a {SubType.__name__}: {err}"
            )
        return sub

    @classmethod
    @contextmanager
    def _get(
        cls,
        ContextDataType: Type[ContextDataType],
        kwargs: Any,
        skip_extra_fields: bool,
    ) -> Generator[ContextDataType, None, None]:
        assert cls is Context or is_dataclass(ContextDataType)
        parent: ContextDataType = cls.current(raises=False)  # type: ignore - we know it is a ContextData
        sub = cls._augment(
            parent, ContextDataType, kwargs, skip_extra_fields=skip_extra_fields
        )
        token = cls._ctx_var.set(sub)
        with sub._on_entering():
            try:
                yield sub
            finally:
                sub._on_exiting()
                cls._ctx_var.reset(token)
                sub._on_exited()

    @classmethod
    @contextmanager
    def get(
        cls, ContextDataType: Type[ContextDataType], **kwargs: Any
    ) -> Generator[ContextDataType, None, None]:
        """
        Create a new context of type ContextDataType.
        The current context must be a subclass of ContextDataType (or ContextDataType must support the
        fields of the current context)
        """
        with cls._get(ContextDataType, kwargs, skip_extra_fields=False) as ctx:
            yield ctx

    @classmethod
    @contextmanager
    def get_as(
        cls, ContextDataType: Type[ContextDataType], **kwargs: Any
    ) -> Generator[ContextDataType, None, None]:
        """
        Create a new context of type ContextDataType.
        Fields of the current context are dropped if not supported
        by ContextDataType.
        """
        with cls._get(ContextDataType, kwargs, skip_extra_fields=True) as ctx:
            yield ctx


@Context._decorate_context_type
class ContextData(Context):
    _depth: int

    @contextmanager
    def _on_entering(self):
        """
        Called right before this context becomes current.
        """
        yield

    def _on_exiting(self):
        """
        Called right before this context exits.
        """
        pass

    def _on_exited(self):
        """
        Called right after this context exited.
        """
        pass



from __future__ import annotations
from typing import Generator, Type, TypeVar, ClassVar, Any, Iterable
from types import GenericAlias

from pathlib import Path
from contextlib import contextmanager
from dataclasses import asdict, is_dataclass

from tgzr.context import Context, ContextData

#
#   Contexts Definition
#

# We will use contexts to provide the traits of an object.
# To do so, we manage a list of known Traits in a registry.
# Each trait is a ContextData so there's always a current one
# and you can ask to elevate it to a sub-trait by creating
# a context of that sub-trait. If it fails, the object doesn't
# have this trait(s)

# Traits registry
# Traits are actually called "DataType" because "Trait" is way
# too pedantic for me ;)

DataType = TypeVar("DataType", bound="Data")


class DataTypeRegistry:
    """
    The DataTypeRegistry knows all the DataType you want.
    It is able to generate one of its DataType from the
    current context (when compatible).
    """

    def __init__(self):
        self._data_types: dict[str, Type[Data]] = {}

    def register(self, cls: Type[DataType]) -> Type[DataType]:
        if not issubclass(cls, Data):
            raise TypeError("Can only register subclasses of Data.")
        self._data_types[cls.__name__] = cls
        return cls

    def get(self, type_name: str) -> Type[Data] | None:
        return self._data_types.get(type_name)


registry = DataTypeRegistry()


@Context.context_type
class Data(ContextData):
    """
    The Data is the root context for our DataTypes (traits).
    It has a registry as class variable, and automatically fills
    it with all the classes  decorated as `@context_type()`

    It is able to dump a DataType to json and to rebuild one from
    json (making it the current one).

    """

    registry: ClassVar[DataTypeRegistry] = DataTypeRegistry()

    @classmethod
    def context_type(cls, class_to_decorate: Type[DataType]) -> Type[DataType]:
        decorated_class = super(Data, cls).context_type(class_to_decorate)
        cls.registry.register(decorated_class)
        return decorated_class

    def to_dict(self) -> dict[str, Any]:
        data = asdict(self)  # type: ignore - we know it's a dataclass
        data["_type"] = self.__class__.__name__
        return data

    @classmethod
    def as_dataclass(cls, DataClass: Type[DataType], data: dict[str, Any]) -> DataType:
        values = {}
        for f in fields(DataClass):  # type: ignore - it's a DataType we know it's a dataclass type
            if isinstance(f.type, GenericAlias) and f.type.__origin__ == list:
                values[f.name] = [
                    cls.as_dataclass(f.type.__args__[0], d2) for d2 in data[f.name]
                ]
            elif not is_dataclass(data[f.name]):
                values[f.name] = data[f.name]
            else:
                values[f.name] = cls.as_dataclass(
                    f.type,  # type: ignore - it's a dataclass because of the earlier "elif no is_dataclass():"
                    data[f.name],
                )

        return DataClass(**values)

    @classmethod
    @contextmanager
    def from_dict(
        cls: Type[DataType], data: dict[str, Any]
    ) -> Generator[DataType, None, None]:
        data = data.copy()
        type_name = data.pop("_type", None)

        if not type_name:
            raise ValueError("Data missing '_type' field for reconstruction.")

        DataType = cls.registry.get(type_name)
        if not DataType:
            raise ValueError(f"Class '{type_name}' not found in Context registry.")

        with cls.get(cls, **data) as d:
            yield d
        # return cls.as_dataclass(DataType, data)


@Data.context_type
class FSData(Data):
    """
    A filesystem DataType, i.e: a Data with a path.

    Not directly usable, you need to use Folder or File
    instead, or make you own subclass which implements
    the `filenames()->list[path]` method.
    """

    path: str

    def filenames(self) -> list[Path]:
        raise NotImplementedError()


@Data.context_type
class Folder(FSData):
    """
    A filesystem Data, i.e: a Data with a path.
    The `filenames()` method returns all the contained files
    or folders.
    """

    def filenames(self) -> list[Path]:
        return list(Path(self.path).iterdir())


@Data.context_type
class File(FSData):
    """
    A File, i.e: a FSData with a name and an extension
    containing a single path.
    """

    name: str
    extension: str

    def filenames(self) -> list[Path]:
        return [(Path(self.path) / self.name).with_suffix(self.extension)]


@Data.context_type
class Sequence(File):
    """
    An ordered Sequence of named files.
    It contains all the files at `{path}/{name}.[{first}->{last}]{.extension}`
    and is represented as `{path}/{name}.{padding}{.extension}`.

    It can iterate over its frame numbers with `iter_fram_numbers()`.
    It can iterate over its `Frame` contexts with `with iter_frames() as frame:`

    You can also get a specific contained Frame:
        with my_sequence.frame(frame_number) as frame:
            assert isinstance(frame, Frame)

    """

    padding: str
    first: int
    last: int

    def iter_frame_numbers(self) -> Iterable[int]:
        for frame_number in range(self.first, self.last + 1):
            yield frame_number

    @contextmanager
    def frame(self, frame_number: int) -> Generator[Frame, None, None]:
        with self.get(Frame, frame_number=frame_number) as frame:
            yield frame

    def iter_frames(self) -> Iterable[Frame]:
        for i in self.iter_frame_numbers():
            with self.frame(i) as frame:
                yield frame

    def filenames(self) -> list[Path]:
        return sum([frame.filenames() for frame in self.iter_frames()], [])


@Data.context_type
class Frame(Sequence):
    """
    A Frame contains one of the files contained in a Sequence.
    The file is identifier by the `frame_number` field.
    """

    frame_number: str

    def filenames(self) -> list[Path]:
        padding_width = self.padding.count("#") * 4 + self.padding.count("@")
        suffix = f".{str(self.frame_number).zfill(padding_width)}{self.extension}"
        return [(Path(self.path) / self.name).with_suffix(suffix)]


def demo_data_traits_context():

    # We create a Sequence context by providing all its fields:
    with Data.get(
        Sequence,
        path="/path/to",
        name="my_shot",
        extension=".exr",
        first=51,
        last=150,
        padding="@",
    ) as seq:
        # It's a Sequence with exactrly 100 frames:
        assert len(list(seq.iter_frame_numbers())) == 100
        assert len(seq.filenames()) == 100

        # It's the current Sequence:
        assert Sequence.current() == seq

        # We can iterate over the sequence, making
        # each frame the current one as we get it.
        # This give you the opportunity to implement
        # methods which can process a frame without
        # providing the frame as argument. It essentially
        # is a dependency injection, which will make your
        # applications easily configurable and extensible
        # while keeping coupling to a minimum (the Frame
        # API but nothing about how to instanciate it).
        for frame in seq.iter_frames():
            # This Frame is the current Frame
            assert Frame.current() is frame

            # We have completion on frame attributes:
            frame.filenames
            frame.iter_frame_numbers
            frame.iter_frames
            frame.first
            frame.last
            frame.path
            frame.name
            frame.padding
            frame.frame_number

        # we can store the state of the context:
        seq_dict = seq.to_dict()
        # We will now use `seq_dict` to "restore" the
        # frame's context.

    # NB: the current (WIP) implementation needs this,
    # but we're working on removing that need:
    del seq_dict["_type"]

    # This is the sequence context dumped into a dict
    # that you could store as json or similar:
    print("->", seq_dict)

    # We can "restore" the original context with
    # `Data.get_as()` or with `Data.get()`
    # Read on to discover the difference!

    # Restoring with `Data.get_as(**dump)`

    # A Sequence is a FSData so
    # we can create a FSData context from the Sequence's dump:
    with Data.get_as(FSData, **seq_dict) as fsdata:
        # It becomes the current context:
        assert Data.current() is fsdata
        assert FSData.current() is fsdata
        # But the current context is just a FSData
        # and cannot be accessed as File
        try:
            File.current()
        except TypeError:
            pass
        else:
            raise AssertionError()
        # and cannot be accessed as Sequence
        try:
            Sequence.current()
        except TypeError:
            pass
        else:
            raise AssertionError()
        # and cannot be accessed as Frame
        try:
            Frame.current()
        except TypeError:
            pass
        else:
            raise AssertionError()

        # It has the exact requested type:
        assert isinstance(fsdata, FSData)
        # With completion for that type:
        fsdata.path

    # A Sequence is a File so
    # we can create a File context from the Sequence's dump:
    with Data.get_as(File, **seq_dict) as file:
        # It becomes the current context:
        assert Data.current() is file
        assert FSData.current() is file
        assert File.current() is file
        # But the current context is just a File
        # and cannot be accessed as Sequence
        try:
            Sequence.current()
        except TypeError:
            pass
        else:
            raise AssertionError()
        # and cannot be accessed as Frame
        try:
            Frame.current()
        except TypeError:
            pass
        else:
            raise AssertionError()

        # It has the exact requested type:
        assert isinstance(file, File)
        # With completion for that type:
        file.path
        file.filenames

    # If you know it's a sequence
    # you can recreate a Sequence context:
    with Data.get_as(Sequence, **seq_dict) as file:
        # It becomes the current context:
        assert Data.current() is file
        assert FSData.current() is file
        assert File.current() is file
        assert Sequence.current() is file
        # And cannot be accessed as Frame
        # since a Sequence is not a Frame
        try:
            Frame.current()
        except TypeError:
            pass
        else:
            raise AssertionError()

    # Restoring with `Data.get(**dump)`

    # Since it's a Sequence dump, we can simply
    # recreate the context by passing the dump
    # as fields for a Sequence context:
    with Data.get(Sequence, **seq_dict) as sequence:
        # It become the current context:
        assert Sequence.current() is sequence
        assert File.current() is sequence
        assert FSData.current() is sequence
        assert Data.current() is sequence
        # It cannot be accessed as Frame
        # since a sequence is not a Frame:
        try:
            Frame.current()
        except TypeError:
            # TypeError: The current context is a <calls `[...].Sequence'>, not a <class '[...].Frame'>
            pass
        else:
            raise AssertionError()

        # It has the exact requested type:
        assert isinstance(sequence, Sequence)
        # With completion on that type:
        sequence.path
        sequence.name
        sequence.filenames
        sequence.padding
        sequence.extension
        sequence.iter_frame_numbers
        sequence.iter_frames

    # Data.get() will raise TypeError if the requested
    # type is lower than the serialized type
    # (or more precisely, if the requested type can't
    # handle the serialized type attribute):
    try:
        with Data.get(File, **seq_dict) as file:
            pass
    except TypeError:
        # Since the File context does not have a 'padding' field:
        # TypeError: __init__() got an unexpected keyword argument 'padding'
        pass
    else:
        raise AssertionError()

    # Data.get() will raise TypeError if the requested
    # type is high than the serialized type
    # (or more precisely, if the serialized type can't
    # provide all the attribute of the requested type):
    try:
        with Data.get(Frame, **seq_dict) as frame:
            pass
    except TypeError:
        # Since the Sequence dumped context did not have a 'frame_number' field:
        # TypeError: __init__() missing 1 required positional argument: 'frame_number'
        pass
    else:
        raise AssertionError()

    # TODO: implement something like Data.get_highest(dump) returning
    # an instannce of the highest class compatible with dump.
    # -> This could be usefull in GUIs but we wouldn't get code completion on
    # it so it's a low priority.


if __name__ == "__main__":
    demo_data_traits_context()



from __future__ import annotations
from typing import Generator, Iterator

from contextlib import contextmanager

from tgzr.context import Context, ContextData

#
#   Contexts Definition
#


@Context.context_type
class ProjectCtx(ContextData):
    project_name: str
    project_id: int
    config_name: str

    def task_depts(self) -> list[str]:
        return ["core", "service", "app"]

    def task_names(self) -> list[str]:
        if self.config_name == "DEV":
            return ["impl", "build", "release"]
        else:
            return ["spec", "test", "report"]

    def user_names(self) -> list[str]:
        if self.config_name == "DEV":
            return ["Bob-Dev", "Alice-Dev"]
        else:
            return ["Bob", "Alice"]

    @contextmanager
    def config(self, config_name: str) -> Generator[ProjectCtx, None, None]:
        with self.get(
            ProjectCtx,
            config_name=config_name,
        ) as p:
            yield p

    @contextmanager
    def user(self, user_name: str) -> Generator[UserCtx, None, None]:
        with self.get(UserCtx, user_name=user_name) as user:
            yield user


@Context.context_type
class UserCtx(ProjectCtx):
    user_name: str

    def iter_tasks(self) -> Iterator[TaskCtx]:
        for task_dept in self.task_depts():
            for name in self.task_names():
                with self.task(task_name=name, task_dept=task_dept) as task:
                    yield task

    @contextmanager
    def task(self, task_name: str, task_dept: str) -> Generator[TaskCtx, None, None]:
        with self.get(TaskCtx, task_name=task_name, task_dept=task_dept) as task:
            yield task


@Context.context_type
class TaskCtx(UserCtx):
    task_name: str
    task_dept: str


def demo_project_context_usage():

    # We can define a function which will work with the current
    # contexts, without the need to pass them as arguments:
    def process_config(config_name: str):
        """
        Prints all the tasks in the current project when
        the given config_name is applied.
        """
        # Get the current project:
        project = ProjectCtx.current()
        # Get the same project with a different config:
        with project.config(config_name=config_name) as project_config:
            # This is now the current Project:
            assert project_config is ProjectCtx.current()
            # The previously current project is not the same because it had a different config:
            assert project is not ProjectCtx.current()
            # But the previously current project has the same id:
            assert project.project_id == ProjectCtx.current().project_id
            # NB: the previous line was 90% written by autocompletion <3

            # There is no current User yet:
            assert UserCtx.current(False) is None

            # Let's iterate over the project's users:
            for user_name in project_config.user_names():
                with project_config.user(user_name=user_name):
                    # There is a current User now:
                    assert UserCtx.current() is not None

                    # Let's iterate over the user's tasks:
                    for task in UserCtx.current().iter_tasks():
                        # The current Task is the one we're using:
                        assert task == TaskCtx.current()
                        # The current User is the current Task,
                        # but since Task is a sub-context of User, you get all the completion for a User
                        assert task == UserCtx.current()
                        # The current Project is the current Task,
                        # but since Task inherit Project you get all the completion for a Project
                        assert task == ProjectCtx.current() and isinstance(
                            task, UserCtx
                        )
                        # The current Context is the current Task
                        # and you get completion for a Task
                        assert task == Context.current() and isinstance(task, TaskCtx)
                        print(
                            "Task:",
                            # All these attributes are avaible in completion:
                            task.project_name,
                            task.project_id,
                            task.config_name,
                            task.user_name,
                            task.task_dept,
                            task.task_name,
                        )

    # We create a new context using the Context.get() classmethod:
    with Context.get(
        ProjectCtx, project_name="MyProject", project_id=1234, config_name=None
    ) as project:
        # The ProjectCtx context became "the current ProjectCtx", and since we
        # gave it `project_name="MyProject"`, we made this Project the current one:
        assert project == ProjectCtx.current()

        # We can call procedure acting on this project without passing the project,
        # the procedure will use the current one:
        process_config(config_name="DEV")
        process_config(config_name="PROD")


if __name__ == "__main__":
    demo_project_context_usage()



from .context import Context, ContextData
from .loggable_context import LoggableContext


# from __future__ import annotations
# from typing import (
#     Type,
#     TypeVar,
#     Literal,
#     Generator,
#     Iterator,
#     overload,
#     Any,
#     ClassVar,
# )
# from types import GenericAlias

# import sys
# from contextlib import contextmanager
# from contextvars import ContextVar
# from dataclasses import dataclass, is_dataclass, fields, asdict, InitVar
# import traceback


# import rich
# import rich.console
# import rich.panel
# import rich.table
# import rich.columns


# T = TypeVar("T")


# class Context:
#     _ctx_var: ContextVar["Context"] = ContextVar("Context")

#     @classmethod
#     def _decorate_context_type(cls, class_to_decorate):
#         # Defaults:
#         dataclass_args = {
#             "frozen": True,
#             "repr": True,
#         }
#         # Optimizations for python 3.10+
#         if sys.version_info >= (3, 10):
#             dataclass_args["slots"] = True
#             dataclass_args["kw_only"] = True
#         decorated = dataclass(class_to_decorate, **dataclass_args)
#         return decorated

#     @classmethod
#     def context_type(cls, class_to_decorate: Type[T]) -> Type[T]:
#         if not issubclass(class_to_decorate, ContextData):
#             raise TypeError("You must inherit ContextData!")
#         for name in class_to_decorate.__annotations__.keys():
#             if name in ("_depth", "_on_entering", "_on_exiting", "_on_exited"):
#                 raise TypeError(
#                     f"You cannot declare the ContextData subclass {class_to_decorate} with a field {name!r}: it is a reserved keyword."
#                 )

#         decorated_class = cls._decorate_context_type(class_to_decorate)
#         return decorated_class

#     @overload
#     @classmethod
#     def current(
#         cls: Type[ContextType], raises: Literal[False]
#     ) -> ContextType | None: ...

#     @overload
#     @classmethod
#     def current(cls: Type[ContextType], raises: Literal[True]) -> ContextType: ...

#     @overload
#     @classmethod
#     def current(cls: Type[ContextType]) -> ContextType: ...

#     @classmethod
#     def current(cls: Type[ContextType], raises: bool = True) -> ContextType | None:
#         ctx = cls._ctx_var.get(None)
#         if ctx is None:
#             if not raises:
#                 return None
#             raise ValueError(f"No context is currently active.")
#         if not isinstance(ctx, cls):
#             if not raises:
#                 return None
#             raise TypeError(f"The current context is a {ctx.__class__}, not a {cls}.")
#         return ctx

#     @classmethod
#     def _augment(
#         cls,
#         parent: ContextData | None,
#         SubType: Type[ContextDataType],
#         kwargs: dict[str, Any],
#         skip_extra_fields: bool,
#     ) -> ContextDataType:
#         sub_kwargs = {}
#         if parent is not None:
#             parent_fields = fields(parent)  # type: ignore - we know it is a dataclass
#             sub_kwargs = {f.name: getattr(parent, f.name) for f in parent_fields}
#         sub_kwargs.update(**kwargs)
#         if skip_extra_fields:
#             filtered = {}
#             for field in fields(SubType):  # type: ignore - we know it is a dataclass
#                 filtered[field.name] = sub_kwargs[field.name]
#             sub_kwargs = filtered
#         depth = sub_kwargs.get("_depth", 0)
#         sub_kwargs["_depth"] = depth + 1
#         # TODO: should we validate all field type here?
#         # NB: I tried to use pydantic.dataclasses.dataclass for that but it's not that straightforward
#         # If we decide that ContextData must validate their fields, we should probably drop dataclasses and use pydantic models...
#         try:
#             sub = SubType(**sub_kwargs)
#         except TypeError as err:
#             raise TypeError(
#                 f"While augmenting a {parent.__class__.__name__} to a {SubType.__name__}: {err}"
#             )
#         return sub

#     @classmethod
#     @contextmanager
#     def _get(
#         cls,
#         ContextDataType: Type[ContextDataType],
#         kwargs: Any,
#         skip_extra_fields: bool,
#     ) -> Generator[ContextDataType, None, None]:
#         assert cls is Context or is_dataclass(ContextDataType)
#         parent: ContextDataType = cls.current(raises=False)  # type: ignore - we know it is a ContextData
#         sub = cls._augment(
#             parent, ContextDataType, kwargs, skip_extra_fields=skip_extra_fields
#         )
#         token = cls._ctx_var.set(sub)
#         with sub._on_entering():
#             try:
#                 yield sub
#             finally:
#                 sub._on_exiting()
#                 cls._ctx_var.reset(token)
#                 sub._on_exited()

#     @classmethod
#     @contextmanager
#     def get(
#         cls, ContextDataType: Type[ContextDataType], **kwargs: Any
#     ) -> Generator[ContextDataType, None, None]:
#         """
#         Create a new context of type ContextDataType.
#         The current context must be a subclass of ContextDataType (or ContextDataType must support the
#         fields of the current context)
#         """
#         with cls._get(ContextDataType, kwargs, skip_extra_fields=False) as ctx:
#             yield ctx

#     @classmethod
#     @contextmanager
#     def get_as(
#         cls, ContextDataType: Type[ContextDataType], **kwargs: Any
#     ) -> Generator[ContextDataType, None, None]:
#         """
#         Create a new context of type ContextDataType.
#         Fields of the current context are dropped if not supported
#         by ContextDataType.
#         """
#         with cls._get(ContextDataType, kwargs, skip_extra_fields=True) as ctx:
#             yield ctx


# ContextType = TypeVar("ContextType", bound=Context)


# @Context._decorate_context_type
# class ContextData(Context):
#     _depth: int

#     @contextmanager
#     def _on_entering(self):
#         """
#         Called right before this context becomes current.
#         """
#         yield

#     def _on_exiting(self):
#         """
#         Called right before this context exits.
#         """
#         pass

#     def _on_exited(self):
#         """
#         Called right after this context exited.
#         """
#         pass


# class RichLogger:
#     def __init__(self, record: bool = False, indent_width=4) -> None:
#         self._console = rich.console.Console(record=record)
#         self._indent = 0
#         self._indent_width = max(2, indent_width)

#     def export_text(self) -> str:
#         return self._console.export_text()

#     def _indented(self, r, lc="│ "):
#         sep = "│"
#         i = f"{{:>{self._indent_width}}}".format(sep)

#         ilc = f"??{{:>{self._indent_width}}}".format(lc)
#         ilc = ((self._indent_width - 1) * " ") + lc
#         col = rich.columns.Columns(
#             [(self._indent // 2) * i + ilc, r],
#             padding=(0, 0),
#             # column_first=True,
#             align="right",
#         )
#         return col

#     def message(self, message: rich.console.RenderableType):
#         r = self._indented(message, lc="├─ ")
#         self._console.print(r)

#     def log_dict(
#         self,
#         d: dict[rich.console.RenderableType, rich.console.RenderableType],
#         title: str | None = None,
#     ):
#         table = rich.table.Table(title=title)
#         for k, v in d.items():
#             table.add_row(str(k), repr(v))
#         self._console.print(self._indented(table))

#     def _section_header(self, title: str):
#         p = rich.panel.Panel(title, expand=False)
#         return p

#     def _section_footer(self):
#         return ""

#     @contextmanager
#     def group(self, title: str):
#         self.message(title)
#         self._indent += 2
#         try:
#             yield
#         finally:
#             # self._indent -= 1
#             self._console.print(self._indented("", lc=f"╰🠆 ✓"))
#             self._indent -= 2

#     @contextmanager
#     def section(self, title: str):
#         w = len(title)
#         i = (self._indent_width - 2) * "─"
#         self._console.print(self._indented("", lc="│╭─" + w * "─" + "─╮"))
#         self._console.print(self._indented("", lc="├┤ " + title + " │"))
#         self._console.print(
#             self._indented("", lc=f"│╰{i}┬" + (w - self._indent_width + 2) * "─" + "─╯")
#         )
#         # self._console.print(rich.padding.Padding.indent(p, (indent * 4) - 2))
#         self._indent += 2
#         try:
#             yield
#         except Exception as err:

#             if footer := self._section_footer():
#                 self._console.print(self._indented("", lc=f"├🠆 {footer}"))

#             self._console.print(self._indented("[red]✘ Error:[/red]", lc=f"├🠆 "))
#             tb = traceback.format_exc()
#             self._console.print(rich.panel.Panel(tb, style="red"))
#             if str(err):
#                 self._console.print(self._indented(f"[red]{err}[/red]", lc=f"├🠆 "))

#         else:
#             if footer := self._section_footer():
#                 self._console.print(self._indented("", lc=f"├🠆 {footer}"))
#             msg = f"[green]🗹[/green] Done {title}"
#             self._console.print(self._indented("", lc=f"╰🠆 {msg}"))
#         finally:
#             self._indent -= 2
#             # self._console.print(self._indented("", lc="├─╯"))


# @Context._decorate_context_type
# class LoggableContext(ContextData):
#     @classmethod
#     def create_logger(cls, record: bool = False) -> RichLogger:
#         return RichLogger(record=record)

#     logger: RichLogger | None

#     def _log_section_summary(self) -> str:
#         return ""

#     @contextmanager
#     def _on_entering(self):
#         if self.logger is not None:
#             with self.logger.section(self._log_section_summary()):
#                 yield
#         return

#     def log(self, message: str):
#         if self.logger is None:
#             return
#         self.logger.message(message)

#     @contextmanager
#     def log_group(self, title: str):
#         if self.logger is None:
#             return
#         with self.logger.group(title):
#             yield

#     @contextmanager
#     def log_section(self, title: str):
#         if self.logger is None:
#             return
#         with self.logger.section(title):
#             yield


# ContextDataType = TypeVar("ContextDataType", bound=ContextData)



from __future__ import annotations

from contextlib import contextmanager
import traceback

import rich
import rich.console
import rich.panel
import rich.table
import rich.columns

from .context import Context, ContextData


class RichLogger:
    def __init__(self, record: bool = False, indent_width=4) -> None:
        self._console = rich.console.Console(record=record)
        self._indent = 0
        self._indent_width = max(2, indent_width)

    def export_text(self) -> str:
        return self._console.export_text()

    def _indented(self, r, lc="│ "):
        sep = "│"
        i = f"{{:>{self._indent_width}}}".format(sep)

        ilc = f"??{{:>{self._indent_width}}}".format(lc)
        ilc = ((self._indent_width - 1) * " ") + lc
        col = rich.columns.Columns(
            [(self._indent // 2) * i + ilc, r],
            padding=(0, 0),
            # column_first=True,
            align="right",
        )
        return col

    def message(self, message: rich.console.RenderableType):
        r = self._indented(message, lc="├─ ")
        self._console.print(r)

    def log_dict(
        self,
        d: dict[rich.console.RenderableType, rich.console.RenderableType],
        title: str | None = None,
    ):
        table = rich.table.Table(title=title)
        for k, v in d.items():
            table.add_row(str(k), repr(v))
        self._console.print(self._indented(table))

    def _section_header(self, title: str):
        p = rich.panel.Panel(title, expand=False)
        return p

    def _section_footer(self):
        return ""

    @contextmanager
    def group(self, title: str):
        self.message(title)
        self._indent += 2
        try:
            yield
        finally:
            # self._indent -= 1
            self._console.print(self._indented("", lc=f"╰🠆 ✓"))
            self._indent -= 2

    @contextmanager
    def section(self, title: str):
        w = len(title)
        i = (self._indent_width - 2) * "─"
        self._console.print(self._indented("", lc="│╭─" + w * "─" + "─╮"))
        self._console.print(self._indented("", lc="├┤ " + title + " │"))
        self._console.print(
            self._indented("", lc=f"│╰{i}┬" + (w - self._indent_width + 2) * "─" + "─╯")
        )
        # self._console.print(rich.padding.Padding.indent(p, (indent * 4) - 2))
        self._indent += 2
        try:
            yield
        except Exception as err:

            if footer := self._section_footer():
                self._console.print(self._indented("", lc=f"├🠆 {footer}"))

            self._console.print(self._indented("[red]✘ Error:[/red]", lc=f"├🠆 "))
            tb = traceback.format_exc()
            self._console.print(rich.panel.Panel(tb, style="red"))
            if str(err):
                self._console.print(self._indented(f"[red]{err}[/red]", lc=f"├🠆 "))

        else:
            if footer := self._section_footer():
                self._console.print(self._indented("", lc=f"├🠆 {footer}"))
            msg = f"[green]🗹[/green] Done {title}"
            self._console.print(self._indented("", lc=f"╰🠆 {msg}"))
        finally:
            self._indent -= 2
            # self._console.print(self._indented("", lc="├─╯"))


@Context._decorate_context_type
class LoggableContext(ContextData):
    @classmethod
    def create_logger(cls, record: bool = False) -> RichLogger:
        return RichLogger(record=record)

    logger: RichLogger | None

    def _log_section_summary(self) -> str:
        return ""

    @contextmanager
    def _on_entering(self):
        if self.logger is not None:
            with self.logger.section(self._log_section_summary()):
                yield
        return

    def log(self, message: str):
        if self.logger is None:
            return
        self.logger.message(message)

    @contextmanager
    def log_group(self, title: str):
        if self.logger is None:
            return
        with self.logger.group(title):
            yield

    @contextmanager
    def log_section(self, title: str):
        if self.logger is None:
            return
        with self.logger.section(title):
            yield



"""
Using Context in a execution graph, and specifically LoggableContext
to get rich logging and report about the execution process.

Here we simulate some kind of "Asset graph" where assets can be executed.
Each Asset type has a specific behavior when executed.
Assets are connected to other Assets input.
The name of the input they connect to is called 'input group' and can be:
- None: default connection
- build: for Builder assets
- edit: for Editor assets
- view: for Viewer assets

The system uses contexts to affect the execution of an upstream Asset from a
downstream asset.
For example:

  ┌───────┐
  │EDITOR ┼─────┐
  └───────┘     │
  ┌───────┐    ┌▼──────┐
  │BUILDER┼────► SCENE │
  └───────┘    └▲──────┘
  ┌───────┐     │
  │PRODUCT┼─────┘
  └───────┘

The SCENE needs to trigger the EDITOR when executed.
If it was never built before, it needs the BUILDER to built a file containing PRODUCT.
EDITOR and BUILDER both need to be generic so we can use it on any SCENE.

So:
- SCENE.execute() create a BuilContext(asset_to_build=self) before calling BUILDER.execute()
- BUILDER.execute() reads inputs from the context's asset_to_build field and call PRODUCT.execute()
- PRODUCT.execute() calls execute on all its inputs
- etc...
"""

from __future__ import annotations
from typing import Generator, Any

from contextlib import contextmanager

from tgzr.context import Context, LoggableContext


#
#   Asset Contexts
#


@Context.context_type
class AssetExcecutionContext(LoggableContext):
    asset: Asset

    def _log_section_summary(self) -> str:
        return f"{self.__class__.__name__} asset={self.asset.name!r} ({self.asset.__class__.__name__})"


@Context.context_type
class BuildContext(AssetExcecutionContext):
    asset: Builder
    asset_to_build: Buildable

    def _log_section_summary(self) -> str:
        summary = AssetExcecutionContext._log_section_summary(self)
        return f"{summary} asset_to_build={self.asset_to_build.name!r} ({self.asset_to_build.__class__.__name__})"


@Context.context_type
class ViewContext(AssetExcecutionContext):
    asset_to_view: Viewable

    def _log_section_summary(self) -> str:
        summary = AssetExcecutionContext._log_section_summary(self)
        return f"{summary} asset_to_view={self.asset_to_view.name!r} ({self.asset_to_view.__class__.__name__})"


@Context.context_type
class EditContext(AssetExcecutionContext):
    asset_to_edit: Viewable

    def _log_section_summary(self) -> str:
        summary = AssetExcecutionContext._log_section_summary(self)
        return f"{summary} asset_to_edit={self.asset_to_edit.name!r} ({self.asset_to_edit.__class__.__name__})"


#
#   Asset Nodes
#
class Asset:
    def __init__(self, name: str):
        self.name = name
        self._deps: dict[str | None, list[Asset]] = dict()

    def add_input(self, asset: Asset, group: str | None = None) -> None:
        try:
            assets = self._deps[group]
        except KeyError:
            assets = []
            self._deps[group] = assets
        assets.append(asset)

    def get_inputs(self, group: str | None = None) -> list[Asset]:
        return self._deps.get(group, [])

    def get_unique_input(self, group: str, required: bool = True):
        inputs = self.get_inputs(group=group)
        if required and not inputs:
            raise Exception(
                f"You need to connect something in {group!r} dependency of asset {self.name!r}!"
            )

        if len(inputs) != 1:
            raise Exception(
                f"You need one and exactly one {group!r} dependency. (got: {inputs})"
            )
        return inputs[0]

    @contextmanager
    def execution_context(self) -> Generator[AssetExcecutionContext, None, None]:
        with Context.get(AssetExcecutionContext, asset=self) as asset_context:
            yield asset_context

    def execute(self) -> None:
        raise NotImplementedError()


class AssetWithParams(Asset):
    def __init__(self, name: str):
        super().__init__(name)
        self._params: dict[str, Any] = dict()

    def set_param(self, param_name: str, param_value: Any):
        self._params[param_name] = param_value

    def set_params(self, **params):
        self._params.update(params)


class Computable(AssetWithParams):
    """
    An Asset returning a value in its 'execute()' method.
    """

    def compute(self) -> Any:
        return None

    def execute(self) -> Any:
        return self.compute()


class Buildable(AssetWithParams):
    """
    An Asset which executes its 'build' input in its context.
    """

    def scene_path(self) -> str:
        return f"/path/to/work/{self.name}/{self.name}.xstage"

    def scene_exists(self) -> bool:
        return False

    def build_if_needed(self):
        if not self.scene_exists():
            AssetExcecutionContext.current().log(
                f"Scene {self.scene_path()} is missing, we need to build it."
            )
            self.build()

    def build(self):
        builder = self.get_unique_input("build")
        with Context.get(BuildContext, asset=builder, asset_to_build=self):
            builder.execute()

    def execute(self):
        self.build_if_needed()


class Viewable(Buildable):
    """
    A Buildable Asset which executes its 'viewer' input in its context after building
    the asset if needed.
    """

    def view(self):
        with self.execution_context() as exec_ctx:
            exec_ctx.log(f"Opening viewer for {self.scene_path()}")
            viewer = self.get_unique_input("view")
            with exec_ctx.get(ViewContext, asset_to_view=self):
                viewer.execute()

    def execute(self):
        super().execute()


class Product(Viewable):
    """
    A Viewable Asset which include itself in the current scene when executed in
    a build context.
    All dependency Products are executed too.
    """

    def execute(self):
        ctx = BuildContext.current(raises=False)
        if ctx is None:
            raise Exception("Cannot execute a Product outside of a build context !")

        inputs = self.get_inputs()
        if inputs:
            with ctx.log_group(f"Executing inputs of {self.name!r}"):
                for input in inputs:
                    input.execute()

        ctx.log(f"Adding Product {self.name} to {ctx.asset_to_build.scene_path()}")


class Editable(Buildable):
    """
    A Buildable Asset which executes its 'edit' input in its context after building
    the asset if needed.
    """

    def edit(self):
        ctx = AssetExcecutionContext.current()
        editor = self.get_unique_input("edit")
        with AssetExcecutionContext.get(EditContext, asset=editor, asset_to_edit=self):
            editor.execute()

    def execute(self):
        super().execute()
        self.edit()


class Env(Computable):
    """
    A Computable Asset returning a dict[str, str] from its executes method.
    All input assets must be Env assets too, and the resulting dict will
    confain the aggregation of all input dicts.

    ! In real Asset context, the inputs order is not 100% assured !
    """

    def as_env(self, asset) -> Env | None:
        if not isinstance(asset, Env):
            return None
        return asset

    def compute(self) -> dict:
        env = dict()
        for input in self.get_inputs():
            if env_input := self.as_env(input):
                data = env_input.execute()
                env.update(data)
        for name, value in self._params.items():
            env[name] = value
        return env

    def execute(self) -> dict[str, str]:
        return {}


class Viewer(Asset):
    """
    An asset which will open the 'scene' from the current Editable asset with
    an defined viewer when executed.
    """

    def launch_viewer(self, scene: str):
        ViewContext.current().log(f"Launching viewer for {scene}")

    def execute(self):
        ctx = ViewContext.current(raises=False)
        if ctx is None:
            raise Exception("Cannot view an asset outside of a ViewContext.")
        with self.execution_context() as exec_ctx:
            exec_ctx.log(f"Viewer executing: {self.name}")
            self.launch_viewer(ctx.asset_to_view.scene_path())


class Editor(Asset):
    """
    An asset which will open the 'scene' from the current Editable asset with
    an defined DCC when executed.
    """

    def launch_editor(self, scene: str):
        EditContext.current().log(f"Launching editor on {scene}")

    def execute(self):
        ctx = EditContext.current(raises=False)
        if ctx is None:
            raise Exception("Cannot edit an asset outside of an EditContext.")
        self.launch_editor(ctx.asset_to_edit.scene_path())


class Builder(Asset):
    """
    An asset which will build the 'scene' from the current Editable asset when
    executed.
    """

    def as_buildable(self, asset) -> Buildable | None:
        if not isinstance(asset, Buildable):
            return None
        return asset

    def as_product(self, asset) -> Product | None:
        if not isinstance(asset, Product):
            return None
        return asset

    def as_env(self, asset) -> Env | None:
        if not isinstance(asset, Env):
            return None
        return asset

    def build(self):
        ctx = BuildContext.current()

        # just for testing, it's ensured in real life:
        assert isinstance(ctx.asset_to_build, Buildable)

        with ctx.log_group(
            f"Executing Products & Envs connected to {ctx.asset_to_build.name!r}"
        ):
            for input_asset in ctx.asset_to_build.get_inputs():
                if product_input := self.as_product(input_asset):
                    with ctx.log_group(f"Executing {product_input.name!r}"):
                        product_input.execute()
                elif env_input := self.as_env(input_asset):
                    with ctx.log_group(f"Executing {env_input.name!r}"):
                        env = env_input.execute()
                        ctx.log(f"Env computed: {env}")
                else:
                    ctx.log(f"Skipping non Product nor Env input: {input_asset.name}")

    def execute(self):
        build_ctx = BuildContext.current(raises=False)
        if build_ctx is None:
            raise Exception("Cannot build an asset outside of a BuildContext.")
        self.build()


class WorkScene(Editable):
    pass


def demo_asset_context():

    # Let's create an Asset graph

    alice = Product("Alice-Actor")
    bob = Product("Bob-Actor")

    kitchen = Product("Kitchen-Set")
    table = Product("Table-Prop")
    chair = Product("Chair-Prop")
    kitchen_casting = Product("Kitcher-Set-Casting")
    kitchen_casting.set_params(nb_chair=4)
    kitchen_casting.add_input(table)
    kitchen_casting.add_input(chair)
    kitchen.add_input(kitchen_casting)

    workscene = WorkScene("Anim")
    workscene.add_input(alice)
    workscene.add_input(bob)
    workscene.add_input(kitchen)

    anim_env = Env("AnimEnv")
    workscene.add_input(anim_env)

    builder = Builder("AnimBuilder")
    workscene.add_input(builder, group="build")

    editor = Editor("AnimDCC")
    workscene.add_input(editor, group="edit")

    viewer = Viewer("AnimViewer")
    workscene.add_input(viewer, group="view")

    picker_tool = Env("PickerTool")
    picker_tool.set_params(
        TOOL_PATH="/path/to/tools/picker_tool",
        PICKER_TOOL_GL="1",
    )
    anim_env.add_input(picker_tool)

    qc_tool = Env("QCTool")
    qc_tool.set_params(
        TOOL_PATH="/path/to/tools/qc_tool",
        QC_TOOL_STRICT="False",
    )
    anim_env.add_input(qc_tool)

    # Let's execute the workscene.
    # It should build itself before launchig its editor
    # Building is done by its builder, which will build all
    # products connected to the scene.
    logger = AssetExcecutionContext.create_logger(record=True)
    with Context.get(AssetExcecutionContext, logger=logger, asset=workscene):
        workscene.execute()

    # print("RECORDED:")
    # rich.print(logger.export_text())


if __name__ == "__main__":
    demo_asset_context()
